/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef PCAP_FRAME_DESCRIPTOR_HXX_
# define PCAP_FRAME_DESCRIPTOR_HXX_

# include <cstring>

# include "frame_descriptor.hh"
# include <wipal/tool/exceptions.hh>
# include <wipal/tool/endianness.hh>

namespace pcapxx
{

  inline
  frame_descriptor::frame_descriptor(const pkthdr* h, const void* b):
    pcap_header_ (new pkthdr (*h)),
    bytes_ (new uint8_t [h->caplen])
  {
    memcpy(bytes_.get(), b, pcap_header()->caplen);
  }

  inline
  frame_descriptor::frame_descriptor(const pkthdr_ptr h, const bytes_ptr b):
    pcap_header_ (h),
    bytes_ (b)
  {
  }

  inline
  frame_descriptor::frame_descriptor(std::istream&	stream,
				     unsigned		idx,
				     bool		swap,
				     int32_t		zone,
				     std::streampos*	pos)
  {
    // PCAP header.
    {
      internals::file_frame_header h;

      if (not stream.read(reinterpret_cast<char*> (&h), sizeof (h)) or
	  stream.gcount() != sizeof (h))
	throw tool::read_error ("Unreadable packet header");
      if (pos)
	*pos += sizeof (h);

      pkthdr* const		p = new pkthdr;

      // Clear padding.  Allows further use of memcmp() to test
      // equality between PCAP headers.
      memset(p, 0, sizeof (*p));

      p->id		= idx;
      p->swapped	= swap;
      p->ts.tv_sec	= tool::extract_long_s(h.ts.tv_sec, swap) + zone;
      p->ts.tv_usec	= tool::extract_long_s(h.ts.tv_usec, swap);
      p->caplen		= tool::extract_long_u(h.caplen, swap);
      p->len		= tool::extract_long_u(h.len, swap);
      pcap_header_.reset(p);
    }

    // Frame body.
    {
      const size_t	caplen = pcap_header()->caplen;
      uint8_t* const	tmp = new uint8_t[caplen];

      if (not stream.read(reinterpret_cast<char*> (tmp), caplen) or
	  stream.gcount() != std::streampos (caplen))
	throw tool::read_error ("Unreadable packet data");
      if (pos)
	*pos += caplen;
      bytes_.reset(tmp);
    }
  }

  inline
  const frame_descriptor::pkthdr_ptr&
  frame_descriptor::pcap_header() const
  {
    return pcap_header_;
  }

  inline
  const frame_descriptor::bytes_ptr&
  frame_descriptor::bytes() const
  {
    return bytes_;
  }

  inline
  std::ostream&
  frame_descriptor::print(std::ostream& os) const
  {
    return os << pcap_header()->id;
  }

  inline
  bool
  frame_descriptor::operator == (const frame_descriptor& rhs) const
  {
    return
      0 == memcmp(pcap_header().get(), rhs.pcap_header().get(), sizeof (pkthdr))
      and
      0 == memcmp(bytes().get(), rhs.bytes().get(), pcap_header()->caplen);
  }

  inline
  bool
  frame_descriptor::operator != (const frame_descriptor& rhs) const
  {
    return not (*this == rhs);
  }

  inline
  std::ostream&
  operator << (std::ostream& os, const frame_descriptor& fd)
  {
    return fd.print(os);
  }

} // End of namespace pcapxx.

#endif // ! PCAP_FRAME_DESCRIPTOR_HXX_
