/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef RTAP_HEADER_HXX_
# define RTAP_HEADER_HXX_

# include "rtap_header.hh"

# include <stdexcept>
# include <cstring>

# include <wipal/tool/endianness.hh>

# define tt_diff(B, E) \
  (reinterpret_cast<const uint8_t*> (E) - reinterpret_cast<const uint8_t*> (B))

namespace rtap
{

  void
  check(const header* h, size_t caplen)
  {
    if (caplen < sizeof (*h))
      throw std::invalid_argument ("Truncated radiotap header");

    if (h->it_version)
      throw std::invalid_argument ("Invalid radiotap version");
  }

  const void*
  first_field(const header* h, size_t caplen)
  {
    const uint32_t* p = &h->it_present;

    while (tt_diff(h, p + 1) < ptrdiff_t (caplen) and
	   tool::extract_little_endian_long_u(*p) & header::ext)
      ++p;
    return tt_diff(h, p + 1) >= ptrdiff_t (caplen) ? 0 : p + 1;
  }

  const void*
  get80211(const header* h, size_t caplen)
  {
    const size_t it_len = tool::extract_little_endian_long_u(h->it_len);

    if (caplen < it_len)
      return 0;

    return reinterpret_cast<const uint8_t*> (h) + it_len;
  }


  bool
  eq_80211(const header* lhs, size_t lhs_caplen,
	   const header* rhs, size_t rhs_caplen)
  {
    check(lhs, lhs_caplen);
    check(rhs, rhs_caplen);

    const void*		lhs_80211 = get80211(lhs, lhs_caplen);
    const void*		rhs_80211 = get80211(rhs, rhs_caplen);

    if (not lhs_80211 or not rhs_80211)
      return false;

    const size_t  lhs_80211_caplen = lhs_caplen - tt_diff(lhs, lhs_80211);
    const size_t  rhs_80211_caplen = rhs_caplen - tt_diff(rhs, rhs_80211);

    if (not lhs_80211_caplen or not rhs_80211_caplen)
      return false;

    return 0 == memcmp(lhs_80211, rhs_80211,
		       std::min(lhs_80211_caplen, rhs_80211_caplen));
  }

  tool::microseconds
  get_tsft(const header* h, size_t caplen, const void* first_field)
  {
    if (not h->it_present & header::tsft)
      throw std::invalid_argument ("Radiotap header has no TSFT field");

    if (caplen < sizeof (*h) + 8 /* TSFT length */)
      throw std::invalid_argument ("Truncated TSFT in radiotap header");

    const uint32_t* const  tsft = static_cast<const uint32_t*> (first_field);
    const uint32_t	   lo = tool::extract_little_endian_long_u(tsft[0]);
    const uint32_t	   hi = tool::extract_little_endian_long_u(tsft[1]);

    return tool::microseconds (hi, lo);
  }

  bool
  eq_time_and_80211(const header* lhs, size_t lhs_caplen,
		    const header* rhs, size_t rhs_caplen,
		    unsigned	precision)
  {
    check(lhs, lhs_caplen);
    check(rhs, rhs_caplen);

    const tool::microseconds l = get_tsft(lhs, lhs_caplen,
					  first_field(lhs, lhs_caplen));
    const tool::microseconds r = get_tsft(rhs, rhs_caplen,
					  first_field(rhs, rhs_caplen));

    return abs(l - r) < precision and eq_80211(lhs, lhs_caplen,
						   rhs, rhs_caplen);
  }

} // End of namespace rtap.

# undef tt_diff

#endif // ! RTAP_HEADER_HXX_
