/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef TOOL_STATIC_BASE_HH_
# define TOOL_STATIC_BASE_HH_

/**
 * @file static_base.hh
 *
 * Base class and tools for static hierarchies.
 *
 * @see static_base
 */

/// Independent helper tools.
namespace tool
{

  /**
   * Static hierarchy terminator.
   *
   * @see static_base
   */
  struct bottom
  {
  };

  /**
   * Undefined virtual type for type traits.
   *
   * @see types
   */
  struct undefined_type
  {
  };

  /**
   * Get an exact type given a type and its Bottom parameter.
   *
   * @code
   *   get_exact<Down, Up>::return_type   = Down
   *   get_exact<bottom, Up>::return_type = Up
   * @endcode
   *
   * @see static_base
   */
  template <class Down, class Up>
  struct get_exact
  {
    typedef Down return_type; ///< The trait's return type.
  };

  /**
   * Template specialization to make get_exact work.
   * @copydoc get_exact
   */
  template <class Up>
  struct get_exact<bottom, Up>
  {
    typedef Up return_type; ///< The trait's return type.
  };

# define tt_get_exact(Down, Up...)			\
  typename ::tool::get_exact<Down, Up>::return_type

  /**
   * Base class of all static hierachies.
   *
   * A static hierarchy is a class hierarchy within which each class
   * knows its exact type thanks to specific template parameters.
   *
   * @e E.g. @code
   *   static_base<cow> : animal<cow> : mammal<cow> : cow
   * @endcode
   *
   * @link static_base @endlink's main role is to provide the exact()
   * and exact_ptr() methods.  These methods allow intermediate
   * classes in the hierarchy to @c static_cast themselves directly to
   * their exact type.  This is perfectly type safe.
   *
   * @e E.g. @code
   *   template <class Exact>
   *   void
   *   animal<Exact>::eat()
   *   {
   *     this->exact().eat_impl(); // Delegate eat() to Exact::eat_impl().
   *   }
   * @endcode
   *
   * In the previous example however, one cannot subclass @c cow
   * because it has no @c Exact parametter.  Use the bottom class to
   * solve this issue:
   *
   * @code
   *   static_base<cow<bottom>> : animal<cow<bottom>> : mammal<cow<bottom>> :
   *     cow<bottom>
   *
   *     or
   *
   *   static_base<highland_cow<bottom>> : animal<highland_cow<bottom>> :
   *     mammal<highland_cow<bottom>> : cow<highland_cow<bottom>> :
   *     highland_cow<bottom>
   * @endcode
   *
   * In this example however, @c cow must provide to @c mammal either
   * @c highland_cow<bottom> either @c cow<bottom>, depending on its
   * template parameter.  This is done using the get_exact trait:
   *
   * @code
   *   template <class Bottom>
   *   struct cow:
   *     public mammal<typename get_exact< Bottom, cow<Bottom> >::return_type>
   *   {
   *     // Inherit mammal<cow<bottom>> if Bottom = bottom,
   *     // inherit mammal<Bottom>, else.
   *
   *     // ...
   *   };
   * @endcode
   *
   * For simple cases, we provide the @c %tt_get_exact() and @c
   * %tt_inherit() preprocessor macros:
   *
   * @code
   *   template <class Bottom>
   *   struct cow: tt_inherit(public mammal, cow<Bottom>)
   *   {
   *     typedef tt_get_exact(Bottom, cow<Bottom>) exact_type;
   *
   *     // ...
   *   };
   * @endcode
   */
  template <class Bottom>
  struct static_base
  {
    /**
     * Exact type for this template instanciation.
     * @see static_base
     */
    typedef tt_get_exact(Bottom, static_base<Bottom>)	exact_type;

    /// @name Downcast methods
    //@{


    /// Safely downcast to the exact type as a const reference.
    const exact_type&	exact() const;

    /// Safely downcast to the exact type as a non-const reference.
    exact_type&		exact();

    /// Safely downcast to the exact type as a const pointer.
    const exact_type*	exact_ptr() const;

    /// Safely downcast to the exact type as a non-const pointer.
    exact_type*		exact_ptr();
    //@}
  };

# define tt_inherit(Class, Parent...) Class<tt_get_exact(Bottom, Parent)>

  /**
   * Virual %types for @link static_base static hierarchies @endlink.
   *
   * In static hierarchies it is often desirable to define <em>virtual
   * %types</em>.  That is, %types that will be defined in subclasses
   * but must exist in the parent class.  (@e e.g. iterable provides
   * an @c iterable::iterator type, but this type is actually defined in
   * @link iterable @endlink's subclasses.)  For technical reasons,
   * one must define virtual %types in external classes.  This is
   * @link types @endlink's role.
   *
   * Declare @c foo 's virtual %types in @c types<foo> :
   *
   * @code
   * template <>
   * struct types<foo>
   * {
   *   typedef bar type1;
   *   typedef baz type2;
   * };
   * @endcode
   *
   * Inside a class, access a virtual type with the @c tt_type()
   * preprocessor macro.
   *
   * @e E.g. @code
   * template <class Bottom>
   * struct iterable : public tt_inherit(public static_base, iterable<Bottom>)
   * {
   *   typedef tt_get_exact(Bottom, iterable<Bottom>)	exact_type;
   *
   *   typedef tt_type(iterator, exact_type)		iterator;
   *
   *   // ...
   * };
   * @endcode
   *
   * @see static_base
   */
  template <class Type>
  struct types
  {
  };

# define tt_type(Type, Base...) typename ::tool::types<Base>::Type

} // End of namespace tool.

# include "static_base.hxx"

#endif // ! TOOL_STATIC_BASE_HH_
