/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef TOOL_TT_OPTIONS_HXX_
# define TOOL_TT_OPTIONS_HXX_

# include <boost/algorithm/string/classification.hpp>
# include <boost/algorithm/string/split.hpp>
# include <boost/lexical_cast.hpp>

# include <cassert>
# include <unistd.h>

# include "tt_options.hh"

namespace tool
{

  namespace opt
  {

    namespace internals
    {

      inline
      void
      usage(const std::string& progname,
	    const std::string& available_options,
	    unsigned ninputs, unsigned noutputs)
      {
	std::cerr << "Invocation:\n\t" << progname << " [options]";
	switch (ninputs)
	  {
	  case 0:					break;
	  case 1:  std::cerr << " <input>";		break;
	  case 2:  std::cerr << " <input-1> <input-2>";	break;
	  default: std::cerr << " <inputs...>";		break;
	  }
	switch (noutputs)
	  {
	  case 0:					break;
	  case 1:  std::cerr << " <output>";		break;
	  default: std::cerr << " <outputs...>";	break;
	  }
	std::cerr << "\n\nOptions:\n";

# define tt_help_print_(OptName) OptName
# define tt_help_print(OptName)						\
	if (available_options.find(tt_opt_  ## OptName) <		\
	    available_options.size())					\
	  std::cerr << tt_help_print_(tt_help_ ## OptName)

	tt_help_print(802cmp);
	tt_help_print(attr);
	tt_help_print(bytescmp);
	tt_help_print(colheads);
	tt_help_print(defaultcmp);
	tt_help_print(errs);
	tt_help_print(help);
	tt_help_print(mapping);
	tt_help_print(endianness);
	tt_help_print(prism_noise);
	tt_help_print(quiet);
	tt_help_print(ignore_ref);
	tt_help_print(usec);
	tt_help_print(timecmp);
	tt_help_print(version);

# undef tt_help_print

	if (available_options.find(tt_opt_attr) < available_options.size())
	  std::cerr << "\nAvailable attributes:\n"
	    "  " tt_attr_tmp "\n"
	    "  " tt_attr_dst_tmp "\n"
	    "  " tt_attr_src_tmp "\n"
	    "  " tt_attr_bss_tmp "\n"
	    "  " tt_attr_src_bss_tmp "\n"
	    "  " tt_attr_seq_tmp "\n"
	    "  " tt_attr_seq_bss_tmp "\n"
	    "  " tt_attr_seq_dst_bss_tmp "\n"
	    "  " tt_attr_seq_src_bss_tmp "\n";
      }

      inline
      void
      version(const std::string& progname)
      {
	std::cerr << progname << "\n"
	  "Trace Tools, version " << PACKAGE_VERSION << '.' << std::endl;
      }

      inline
      std::pair<unsigned, unsigned>
      make_reference(const std::string& s)
      {
	static const std::string err = ("Invalid reference frame format. "
					"Format is ``<unsigned>-<unsigned>''");

	std::vector<std::string>	v;

	boost::split(v, s, boost::is_any_of("-"));
	if (v.size() != 2)
	  throw std::invalid_argument (err);
	try
	  {
	    return std::make_pair(boost::lexical_cast<unsigned> (v[0]),
				  boost::lexical_cast<unsigned> (v[1]));
	  }
	catch (const boost::bad_lexical_cast&)
	  {
	    throw std::invalid_argument (err);
	  }
      }

    }

    inline
    options
    parse_args(int argc, char* const* argv,
	       const std::string& available_options,
	       int ninputs, int noutputs)
    {
      const std::string optstring =
	available_options + tt_opt_help tt_opt_version;

      options	r;
      int	o;

      r.mapping = tt_mapping_default;
      r.attributes = tt_attr_default;
      r.endianness = endian::guess;

      r.cmp = options::cmp_default;
      r.prism_noise_means_error = false;
      r.print_errors = false;
      r.print_frame_indices = true;
      r.print_column_headers = isatty(STDOUT_FILENO);
      r.print_microseconds = false;
      r.print_summary = true;

      while ((o = getopt(argc, argv, optstring.c_str())) != -1)
	switch (o)
	  {
	  case '8':
	    r.cmp = options::cmp_802;
	    break;

	  case 'a':
	    r.attributes = optarg;
	    break;

	  case 'b':
	    r.cmp = options::cmp_bytes;
	    break;

	  case 'c':
	    r.print_column_headers = false;
	    break;

	  case 'C':
	    r.print_column_headers = true;
	    break;

	  case 'd':
	    r.cmp = options::cmp_default;
	    break;

	  case 'e':
	    r.print_errors = false;
	    break;

	  case 'E':
	    r.print_errors = true;
	    break;

	  case 'h':
	    internals::usage(argv[0], optstring, ninputs, noutputs);
	    exit(EXIT_SUCCESS);
	    break;

	  case 'i':
	    r.print_frame_indices = false;
	    break;

	  case 'I':
	    r.print_frame_indices = true;
	    break;

	  case 'm':
	    r.mapping = optarg;
	    break;

	  case 'n':
	    r.endianness = endian::little;
	    break;

	  case 'N':
	    r.endianness = endian::big;
	    break;

	  case 'p':
	    r.prism_noise_means_error = false;
	    break;

	  case 'P':
	    r.prism_noise_means_error = true;
	    break;

	  case 'q':
	    r.print_errors = false;
	    r.print_frame_indices = false;
	    r.print_column_headers = false;
	    r.print_microseconds = false;
	    r.print_summary = false;
	    break;

	  case 'r':
	    try
	      {
		r.ref_blacklist.insert(internals::make_reference(optarg));
	      }
	    catch (const std::invalid_argument& i)
	      {
		std::cerr << "ERROR: " << i.what() << '.' << std::endl;
		exit(EXIT_FAILURE);
	      }
	    break;

	  case 't':
	    r.cmp = options::cmp_time;
	    break;

	  case 'u':
	    r.print_microseconds = false;
	    break;

	  case 'U':
	      r.print_microseconds = true;
	      break;

	  case 'v':
	    internals::version(argv[0]);
	    exit(EXIT_SUCCESS);

	  default: /* '?' */
	    internals::usage(argv[0], optstring, ninputs, noutputs);
	    exit(EXIT_FAILURE);
	  }

      assert(ninputs  != options::arbitrary_number_of_inputs or
	     noutputs != options::arbitrary_number_of_outputs);

      if (ninputs == options::arbitrary_number_of_inputs)
	ninputs = argc - optind - noutputs;
      if (noutputs == options::arbitrary_number_of_outputs)
	noutputs = argc - optind - ninputs;
      if (ninputs < 0 or noutputs < 0 or optind + ninputs + noutputs != argc)
	{
	  std::cerr << "ERROR: Wrong number of arguments." << std::endl;
	  internals::usage(argv[0], optstring, ninputs, noutputs);
	  exit(EXIT_FAILURE);
	}
      for (int i = 0; i < ninputs; ++i)
	r.input.push_back(argv[optind + i]);
      for (int i = 0; i < noutputs; ++i)
	r.output.push_back(argv[optind + ninputs + i]);

      return r;
    }

  } // End of namespace tool::opt.

} // End of namespace tool.

#endif // TOOL_TT_OPTIONS_HXX_
