/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef TOOL_VALUED_ITERATOR_HH_
# define TOOL_VALUED_ITERATOR_HH_

# include <boost/optional.hpp>

# include <wipal/tool/iterator.hh>

namespace tool
{

  /*---------------------.
  | Forward declarations |
  `---------------------*/

  template <class B>
  struct valued_iterator;


  /*--------------.
  | Virtual types |
  `--------------*/

  /// Virtual %types for valued_iterator. @see types
  template <class B>
  struct types< valued_iterator<B> >
  {
    /// The iterator's value type.
    typedef undefined_type	value_type;

    /// Type of the iterable structure the iterator is obtained from.
    typedef undefined_type	iterable_type;
  };

  /*-------------.
  | Declarations |
  `-------------*/

  /**
   * Iterator which embeds its value.
   *
   * valued_iterator represents an iterator that embeds its value
   * inside a @c boost::optional<value_type> attribute.  Subclasses
   * access this value using value().  valued_iterator implements the
   * get() and get_ptr() methods which are required by iterator.
   * These methods return the embeded value (with undefined results if
   * the value is @c boost::none_t).  Implement the @c increment() and
   * @c equal() methods to subclass it.  @c increment() should either
   * store the next element into value(), either store boost::none_t
   * to indicate end of iteration.
   *
   * Subclasses must also declare the value_type @link types virtual
   * type @endlink.
   *
   * @see iterator
   */
  template <class Bottom>
  struct valued_iterator: tt_inherit(public iterator, valued_iterator<Bottom>)
  {

    /// @name Static hierarchies related types
    //@{
    typedef tt_get_exact(Bottom, valued_iterator<Bottom>)	exact_type;
    typedef iterator<exact_type>				super_type;
    typedef tt_type(value_type, exact_type)			value_type;
    //@}

    /// @name Indirection as a reference
    //@{
    const value_type&		get() const;
    value_type&			get();
    //@}

    /// @name Indirection as a pointer
    //@{
    const value_type*		get_ptr() const;
    value_type*			get_ptr();
    //@}

  protected:

    /// Protected destructor: valued_iterator is abstract.
    ~valued_iterator();

    /// Type of the embedded value.
    typedef typename boost::optional<value_type>	optional_value;

    /// @name Access to the iterator's embedded value
    //@{
    const optional_value&	value() const;
    optional_value&		value();
    //@}

  private:

    /// Current value, or @c boost::none_t.
    optional_value		v_;

  };

} // End of namespace tool.

# include "valued_iterator.hxx"

#endif // ! TOOL_VALUED_ITERATOR_HH_
