/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef WIFI_FRAME_FILTER_FAST_INTERSECTOR_HXX_
# define WIFI_FRAME_FILTER_FAST_INTERSECTOR_HXX_

# include <iostream>
# include <boost/foreach.hpp>

# include "fast_intersector.hh"

# include <wipal/wifi/frame/filter/non_noisy_prism.hh>
# include <wipal/wifi/frame/filter/uniquely_identifiable.hh>

namespace wifi
{
  namespace frame
  {
    namespace filter
    {

      namespace internals
      {

	template <class I1, class I2, class B1, class B2>
	fast_intersector_iterator<I1, I2, B1, B2>::
	fast_intersector_iterator(const iterable_type& i, bool end):
	  super_type (),
	  iterable_ (&i),
	  next_ (end ? i.last_ : i.first_)
# ifdef ENABLE_INFO
	, rhs_total_ (0),
	  ref_total_ (0)
# endif // ENABLE_INFO
	{
	  if (not end)
	    increment();
	}

	template <class I1, class I2, class B1, class B2>
	bool
	fast_intersector_iterator<I1, I2, B1, B2>::
	equal(const fast_intersector_iterator& rhs) const
	{
	  if (not rhs.value())
	    return not this->value();
	  return next_ == rhs.next_;
	}

	template <class I1, class I2, class B1, class B2>
	void
	fast_intersector_iterator<I1, I2, B1, B2>::increment()
	{
	  const I2&		last = this->iterable_->last_;
	  bool			found (false);

	  while (next_ != last and not found)
	    {
	      found = check();
	      ++next_;
# ifdef ENABLE_INFO
	      ++rhs_total_;
# endif // ENABLE_INFO
	    }

	  if (not found)
	    {
	      this->value() = lookahead_;
	      lookahead_ = boost::none_t ();
	    }

# ifdef ENABLE_INFO
	  if (not found and not this->value())
	    {
	      const unsigned	lhs_total = this->iterable_->arrivals_.size();

	      std::cerr << "INFO: Trace 2 has " << rhs_total_
			<< " unique frames.\n"
			<< "INFO: " << ref_total_ << " reference frames.\n"
			<< "INFO: " << (ref_total_ * 100.f / lhs_total)
			<< "% of trace 1's unique frames.\n"
			<< "INFO: " << (ref_total_ * 100.f / rhs_total_)
			<< "% of trace 2's unique frames." << std::endl;
	    }
	  else
	    ++ref_total_;
# endif // ENABLE_INFO
	}

	template <class I1, class I2, class B1, class B2>
	bool
	fast_intersector_iterator<I1, I2, B1, B2>::check()
	{
	  typedef typename iterable_type::arrivals_type		arrivals_type;
	  typedef typename arrivals_type::const_iterator	iterator;

	  const arrivals_type&	arrivals = iterable_->arrivals_;
	  const iterator	i = arrivals.find(*next_);

	  if (i == arrivals.end())
	    return false;

	  if (not lookahead_)
	    {
	      // First reference we encounter.  Delay it so we can
	      // filter it out in case it is dubious.

	      lookahead_ = value_type (*i, *next_);
	      return false;
	    }
	  assert(lookahead_);

	  value_type&			lookahead = lookahead_.get();
	  const tool::microseconds	dt1 (i->microseconds() -
					     lookahead.first.microseconds());
	  const tool::microseconds	dt2 (next_->microseconds() -
					     lookahead.second.microseconds());
	  const tool::microseconds	drift (dt1 - dt2);

	  if (drift < -drift_warn_threshold or drift_warn_threshold < drift)
	    {
	      std::cerr << "WARNING: Dubious drift of " << (drift / 1000000.f)
			<< "s for "
			<< i->frame_id() << '-' << next_->frame_id();
	      if (this->value())
		{
		  // General case.  The dubious frame is the current one.

		  std::cerr << " (ignoring reference)." << std::endl;
		  return false;
		}
	      else
		{
		  // We are on the second frame.  Consider the first frame
		  // as dubious, and delay this one.

		  std::cerr << ".\n" "WARNING: Ignoring the potentially "
		               "invalid previous reference "
			    << lookahead.first.frame_id() << '-'
			    << lookahead.second.frame_id() << '.' << std::endl;

		  lookahead = value_type (*i, *next_);
		  return false;
		}
	    }

	  if (i->microseconds() < lookahead.first.microseconds())
	    std::cerr << "WARNING: Reference " << i->frame_id() << '-'
		      << next_->frame_id() << " anterior to "
		      << lookahead.first.frame_id() << '-'
		      << lookahead.second.frame_id() << '.' << std::endl;
	  else if (dt1 > dt_warn_threshold or dt2 > dt_warn_threshold)
	    std::cerr << "WARNING: " << (dt1 / 1000000.f) << "s and "
		      << (dt2 / 1000000.f) << "s between "
		      << lookahead.first.frame_id() << '-'
		      << lookahead.second.frame_id() << " and "
		      << i->frame_id() << '-' << next_->frame_id()
		      << '.' << std::endl;

	  this->value() = lookahead;
	  lookahead = value_type (*i, *next_);
	  return true;
	}

      } // End of namespace internals.


      template <class I1, class I2, class B>
      fast_intersector<I1, I2, B>::
      fast_intersector(const I1& first1, const I1& last1,
		       const I2& first2, const I2& last2):
	first_ (first2),
	last_ (last2)
# ifndef DONT_USE_HASH
	, arrivals_ (hash_bucket_count)
# endif
      {
	arrivals_type	collisions;

	for (I1 i = first1; i != last1; ++i)
	  {
	    typedef typename arrivals_type::iterator	iterator;

	    std::pair<iterator, bool> r = arrivals_.insert(*i);

	    if (not r.second)
	      {
		collisions.insert(*i);
		std::cerr << "WARNING: Unique frame " << i->frame_id()
			  << " actually a duplicate of "
			  << r.first->frame_id() << " (ignoring both)."
			  << std::endl;
	      }
	  }

	BOOST_FOREACH(const typename I1::value_type& v, collisions)
	  arrivals_.erase(v);

# ifdef ENABLE_INFO
	std::cerr << "INFO: Trace 1's unique frames successfuly loaded.\n"
		  << "INFO: Trace 1 has " << arrivals_.size()
		  << " unique frames." << std::endl;
# endif // ENABLE_INFO

      }

    } // End of namespace wifi::frame::filter.

  } // End of namespace wifi::frame.

} // End of namespace wifi.

#endif // ! WIFI_FRAME_FILTER_FAST_INTERSECTOR_HXX_
