/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef WIFI_FRAME_FILTER_MERGE_HXX_
# define WIFI_FRAME_FILTER_MERGE_HXX_

# include "merge.hh"

# include <wipal/wifi/frame/filter/time_adjuster.hh>

namespace wifi
{
  namespace frame
  {
    namespace filter
    {

      namespace internals
      {

	inline
	equal<prism::header>::equal(const pcapxx::frame_descriptor& lhs,
				    const tool::endian::endianness& phy_end):
	  lhs_ (lhs),
	  phy_end_ (phy_end)
	{
	}

	inline
	bool
	equal<prism::header>::
	operator == (const pcapxx::frame_descriptor& rhs) const
	{
	  using tool::endian::need_swap;

	  typedef prism::header		bt;
	  typedef pcapxx::pkthdr	pt;

	  const pt* const lp = lhs_.pcap_header().get();
	  const bool	  ls = need_swap(phy_end_, lp->swapped);
	  const bt* const lb = reinterpret_cast<const bt*> (lhs_.bytes().get());

	  const pt* const rp = rhs.pcap_header().get();
	  const bool	  rs = need_swap(phy_end_, rp->swapped);
	  const bt* const rb = reinterpret_cast<const bt*> (rhs.bytes().get());

	  return eq_time_and_80211(lb, lp->caplen, ls,
				   rb, rp->caplen, rs);
	}

	template <class HT>
	bool
	operator == (const pcapxx::frame_descriptor& lhs, const equal<HT>& rhs)
	{
	  return rhs == lhs;
	}

	template <class HT, template <class, class, class, class> class Merger,
		  class I, class F>
	struct provide_merge
	{
	  provide_merge(const I&			first2,
			const I&			last2,
			F&				func,
			bool				filter_prism,
			tool::endian::endianness	phy_end):
	    first2_ (first2),
	    last2_ (last2),
	    func_ (func),
	    filter_prism_ (filter_prism),
	    phy_end_ (phy_end)
	  {
	  }

	  template <class TimeAdjuster>
	  void
	  operator () (const TimeAdjuster& a)
	  {
	    if (filter_prism_)
	      {
		typedef non_noisy_prism<I>			nnp;
		typedef typename nnp::const_iterator		n_iterator;
		typedef microseconds_stamper<n_iterator, HT>	us_stamper;
		typedef typename us_stamper::const_iterator	u_iterator;
		typedef typename TimeAdjuster::const_iterator	a_iterator;
		typedef
		  Merger<HT, a_iterator, u_iterator, tool::bottom> merger;

		nnp		n (first2_, last2_);
		us_stamper	s (n.begin(), n.end(), phy_end_);
		merger		m (a.begin(), a.end(),
				   s.begin(), s.end(), phy_end_);

		func_(m);
	      }
	    else
	      {
		typedef microseconds_stamper<I, HT>		us_stamper;
		typedef typename us_stamper::const_iterator	u_iterator;
		typedef typename TimeAdjuster::const_iterator	a_iterator;
		typedef
		  Merger<HT, a_iterator, u_iterator, tool::bottom> merger;

		us_stamper	s (first2_, last2_, phy_end_);
		merger		m (a.begin(), a.end(),
				   s.begin(), s.end(), phy_end_);

		func_(m);
	      }
	  }

	private:
	  const I&			first2_;
	  const I&			last2_;
	  F&				func_;
	  bool				filter_prism_;
	  tool::endian::endianness	phy_end_;
	};

      } // End of namespace wifi::frame::filter::internals.

      template <class U, class HT,
		template <class, class, class> class Int,
		template <class, class, class, class> class Merger,
		class I1, class I2, class F, class BL>
      void
      provide_merge(const I1& first1, const I1&	last1,
		    const I2& first2, const I2&	last2,
		    addr_mapping&		mapping,
		    F&				func,
		    bool			filter_prism,
		    tool::endian::endianness	phy_end,
		    const BL&			blist)
      {
	internals::provide_merge<HT, Merger, I1, F> func2 (first2, last2,
							   func,
							   filter_prism,
							   phy_end);

	provide_time_adjuster<U, HT, Int>(first1, last1,
					  first2, last2,
					  mapping, func2,
					  filter_prism, phy_end, blist);
      }

    } // End of namespace wifi::frame::filter.

  } // End of namespace wifi::frame.

} // End of namespace wifi.

#endif // ! WIFI_FRAME_FILTER_MERGE_HXX_
