/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef WIFI_FRAME_FILTER_TIME_ADJUSTER_HH_
# define WIFI_FRAME_FILTER_TIME_ADJUSTER_HH_

# include <boost/optional.hpp>

# include <wipal/tool/microseconds_stamp.hh>
# include <wipal/pcap/frame_descriptor.hh>
# include <wipal/tool/valued_iterator.hh>
# include <wipal/tool/iterable.hh>
# include <wipal/wifi/time_converter.hh>
# include <wipal/wifi/addr_mapping.hh>

# include "time_adjuster_fwd.hh"

namespace tool
{

  template <class I1, class I2, class HT, class B1, class B2>
  struct types< wifi::frame::filter::internals::
		time_adjuster_iterator<I1, I2, HT, B1, B2> >
  {
    typedef wifi::frame::filter::time_adjuster<I1, I2, HT, B1>	iterable_type;
    typedef tool::microseconds_stamp<pcapxx::frame_descriptor>	value_type;
  };

  template <class I1, class I2, class HT, class B>
  struct types< wifi::frame::filter::time_adjuster<I1, I2, HT, B> >
  {
    typedef wifi::frame::filter::internals::
      time_adjuster_iterator<I1, I2, HT, B, bottom>
      iterator;
  };

} // End of namespace tool.

namespace wifi
{
  namespace frame
  {
    namespace filter
    {

      namespace internals
      {

	template <class InputIterator1,
		  class InputIterator2,
		  class HeaderType,
		  class B,
		  class Bottom = tool::bottom>
	struct time_adjuster_iterator:
	  tt_inherit(public tool::valued_iterator,
		     time_adjuster_iterator<InputIterator1, InputIterator2,
					    HeaderType, B, Bottom>)
	{
	  /// @name Static hierarchies related types.
	  //@{
	  typedef tt_get_exact(Bottom, time_adjuster_iterator<InputIterator1,
							      InputIterator2,
							      HeaderType,
							      B, Bottom>)
							exact_type;
	  typedef tool::valued_iterator<exact_type>	super_type;
	  typedef tt_type(value_type, exact_type)	value_type;
	  typedef tt_type(iterable_type, exact_type)	iterable_type;
	  //@}

	  /// tool::iterable compatible constructor.
	  time_adjuster_iterator(const iterable_type& i, bool end);

	  /// @name Abstract method implementations
	  //@{
	  bool		equal(const time_adjuster_iterator& rhs) const;
	  void		increment();
	  //@}

	private:
	  typedef typename InputIterator2::value_type::coefs_type coefs_type;

	  void		update_coefs();

	  const iterable_type*		iterable_;
	  boost::optional<coefs_type>	coefs_;

	  InputIterator1		next1_;
	  InputIterator2		next2_;
	};

	template <class HeaderType>
	time_converter<>
	time_reference(const pcapxx::pkthdr&	pcap_header,
		       const HeaderType*	bytes,
		       tool::endian::endianness	phy_end);

	template <class HeaderType>
	void
	adjust(HeaderType* bytes, size_t caplen, bool swap,
	       const time_converter<>::time_values& time);

      } // End of namespace internals.

      template <class InputIterator1,
		class InputIterator2,
		class HeaderType,
		class Bottom = tool::bottom>
      struct time_adjuster: tt_inherit(public tool::iterable,
				       time_adjuster<InputIterator1,
						     InputIterator2,
						     HeaderType,
						     Bottom>)
      {
	time_adjuster(const InputIterator1&		first1,
		      const InputIterator1&		last1,
		      const InputIterator2&		first2,
		      const InputIterator2&		last2,
		      const pcapxx::frame_descriptor&	time_reference,
		      tool::endian::endianness		phy_end);

      private:
	time_converter<>		convert_;
	tool::endian::endianness	phy_end_;

	InputIterator1			first1_;
	InputIterator1			last1_;
	InputIterator2			first2_;
	InputIterator2			last2_;

	friend class internals::time_adjuster_iterator<InputIterator1,
						       InputIterator2,
						       HeaderType,
						       Bottom, tool::bottom>;
      };

      /**
       * Build a time_adjuster filter from iterators and pass it as
       * argument to a functor.
       *
       * If @c filter_prism is true, filter elements with @c
       * non_noisy_prism. @c *first and @c *last should be convertible
       * to @c pcapxx::frame_descriptor.
       *
       * @c phy_end indicates the PHY headers' endianness.
       */
      template <class U, class HT, template <class, class, class> class Int,
		class I1, class I2, class F, class BL>
      void
      provide_time_adjuster(const I1& first1, const I1&	last1,
			    const I2& first2, const I2&	last2,
			    addr_mapping&		mapping,
			    F&				func,
			    bool			filter_prism,
			    tool::endian::endianness	phy_end,
			    const BL&			blist);

    } // End of namespace wifi::frame::filter.

  } // End of namespace wifi::frame.

} // End of namespace wifi.

# include "time_adjuster.hxx"

#endif // ! WIFI_FRAME_FILTER_TIME_ADJUSTER_HH_
