/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */
#ifndef WIFI_FRAME_FILTER_UNIQUELY_IDENTIFIABLE_HH_
# define WIFI_FRAME_FILTER_UNIQUELY_IDENTIFIABLE_HH_

# include <wipal/tool/endianness.hh>
# include <wipal/tool/microseconds_stamp.hh>
# include <wipal/tool/valued_iterator.hh>
# include <wipal/wifi/mactime_tracker.hh>
# include <wipal/tool/iterable.hh>
# include <wipal/wifi/addr_mapping.hh>

# include "uniquely_identifiable_fwd.hh"

/*--------------.
| Virtual types |
`--------------*/

namespace tool
{

  /**
   * Virtual %types for
   * wifi::frame::filter::internals::uniquely_identifiable_iterator.
   *
   * @see types
   */
  template <class U, class I, class H, class B1, class B2>
  struct types< wifi::frame::filter::internals::
		uniquely_identifiable_iterator<U, I, H, B1, B2> >
  {
    typedef
      wifi::frame::filter::uniquely_identifiable<U, I, H, B1>	iterable_type;
    typedef tool::microseconds_stamp<U>				value_type;
  };

  /**
   * Virtual %types for wifi::frame::filter::uniquely_identifiable.
   *
   * @see types
   */
  template <class U, class I, class H, class B>
  struct types< wifi::frame::filter::uniquely_identifiable<U, I, H, B> >
  {
    typedef wifi::frame::filter::internals::
      uniquely_identifiable_iterator<U, I, H, B, bottom>
      iterator;
  };

} // End of namespace tool.

/*-------------.
| Declarations |
`-------------*/

namespace wifi
{
  namespace frame
  {

    /**
     * Various %frame filters.
     *
     * A %filter is an iterable object which filters out various
     * elements from a given iterator.
     */
    namespace filter
    {

      /// Internal compounds - not designed for users.
      namespace internals
      {

	/// Iterator over uniquely identifiable frames.
	template <class U, class I, class H, class B, class Bottom>
	struct uniquely_identifiable_iterator:
	  tt_inherit(public tool::valued_iterator,
		     uniquely_identifiable_iterator<U, I, H, B, Bottom>)
	{

	  /// @name Static hierarchies related types
	  //@{
	  typedef tt_get_exact(Bottom,
			       uniquely_identifiable_iterator<U, I, H,
							      B, Bottom>)
								exact_type;
	  typedef tool::valued_iterator<exact_type>		super_type;
	  typedef tt_type(value_type, exact_type)		value_type;
	  typedef tt_type(iterable_type, exact_type)		iterable_type;
	  //@}

	  /// tool::iterable compatible constructor.
	  uniquely_identifiable_iterator(const iterable_type& uif, bool end);

	  /// @name Abstract method implementations
	  //@{
	  bool		equal(const uniquely_identifiable_iterator& rhs) const;
	  void		increment();
	  //@}

	private:
	  const iterable_type*	iterable_;
	  I			next_;
	  mactime_tracker<>	mactime_;
	};

      } // End of namespace wifi::frame::filter::internals.


      /// Filter uniquely identifiable frames out of PCAP frames.
      template <class UniqueId,
		class InputIterator,
		class HeaderType,
		class Bottom = tool::bottom>
      struct uniquely_identifiable:
	tt_inherit(public tool::iterable,
		   uniquely_identifiable<UniqueId, InputIterator, HeaderType,
					 Bottom>)
      {
	/**
	 * Construct from two iterators and a @link addr_mapping
	 * address mapping @endlink.
	 */
	uniquely_identifiable(const InputIterator&	begin,
			      const InputIterator&	end,
			      addr_mapping&		map,
			      tool::endian::endianness	phy_end);

      private:
	InputIterator			begin_;
	InputIterator			end_;
	addr_mapping*			mapping_;
	tool::endian::endianness	phy_end_;

	friend
	  class internals::uniquely_identifiable_iterator<UniqueId,
							  InputIterator,
							  HeaderType,
							  Bottom,
							  tool::bottom>;
      };

      /**
       * Apply @c func to all uniquely identifiable frames from @c
       * first to @c last.
       *
       * If @c filter_prism is true, filter elements with @c
       * non_noisy_prism. @c *first and @c *last should be convertible
       * to @c pcapxx::frame_descriptor.
       *
       * @c phy_end indicates the PHY headers' endianness.
       */
      template <class UniqueId, class HeaderType, class I, class F>
      void
      for_each_uniquely_identifiable(const I&			first,
				     const I&			last,
				     addr_mapping&		mapping,
				     F&				func,
				     bool			filter_prism,
				     tool::endian::endianness	phy_end);

      /// Const version of @c for_each_uniquely_identifiable.
      template <class UniqueId, class HeaderType, class I, class F>
      void
      for_each_uniquely_identifiable(const I&			first,
				     const I&			last,
				     addr_mapping&		mapping,
				     const F&			func,
				     bool			filter_prism,
				     tool::endian::endianness	phy_end);

    } // End of namespace wifi::frame::filter.

  } // End of namespace wifi::frame.

} // End of namespace wifi.

# include "uniquely_identifiable.hxx"

#endif // ! WIFI_FRAME_FILTER_UNIQUELY_IDENTIFIABLE_HH_
