/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007, 2008  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */

#include <iostream>
#include <stdexcept>
#include <cstdlib>

#include <gmpxx.h>

#include <boost/format.hpp>

#include "config.hh"

#include <wipal/tool/tt_options.hh>
#include <wipal/wifi/frame/filter/linear_regression_synchronizer.hh>
#include <wipal/wifi/frame/filter/intersector.hh>
#include <wipal/wifi/frame/filter/fast_intersector.hh>
#include <wipal/wifi/frame/unique_id/seqctl_bssid_timestamp.hh>
#include <wipal/pcap/list_of_traces.hh>
#include <wipal/wifi/addr_mapping.hh>

namespace
{

  struct sync_unique_frames_dumper
  {

    sync_unique_frames_dumper(const tool::opt::options& options):
      options_ (options),
      coefs_ ("%10s %15s"),
      indices_ (" %10s %10s"),
      msecs_   (" %15s %15s"),
      errors_  (" %10s")
    {
    }

    void	print_colheads()
    {
      std::cout << (coefs_ % "a" % "b");
      if (options_.print_frame_indices)
	std::cout << (indices_ % "frame 1" % "frame 2");
      if (options_.print_microseconds)
	std::cout << (msecs_ % "time 1 (us)" % "time 2 (us)");
      if (options_.print_errors)
	std::cout << (errors_ % "error");

      std::cout << '\n'
		<< (coefs_ % "=" % "=");
      if (options_.print_frame_indices)
	std::cout << (indices_ % "=======" % "=======");
      if (options_.print_microseconds)
	std::cout << (msecs_ % "===========" % "===========");
      if (options_.print_errors)
	std::cout << (errors_ % "=====");
      std::cout << std::endl;
    }

    template <class LRSync>
    void	operator () (const LRSync& s)
    {
      typedef typename LRSync::const_iterator	iterator;

      if (options_.print_column_headers)
	print_colheads();

      const iterator	end = s.end();

      for (iterator j = s.begin(); j != end; ++j)
	{
	  std::cout << (coefs_ % j->coefs.first % j->coefs.second);
	  if (options_.print_frame_indices)
	    std::cout << (indices_ % j->first % j->second);
	  if (options_.print_microseconds)
	    std::cout << (msecs_ % j->first.microseconds()
			         % j->second.microseconds());
	  if (options_.print_errors)
	    std::cout << (errors_ % (j->first.microseconds() * j->coefs.first +
				     j->coefs.second -
				     j->second.microseconds()));
	  std::cout << '\n';
	}
    }

  private:
    const tool::opt::options&	options_;
    boost::format		coefs_;
    boost::format		indices_;
    boost::format		msecs_;
    boost::format		errors_;
  };

} // End of anonymous namespace.

int
main(int argc, char *const argv[])
{
  tool::opt::options	options = tool::opt::parse_args(argc, argv,
							tt_opt_colheads
							tt_opt_errs
							tt_opt_frm_idx
							tt_opt_mapping
							tt_opt_endianness
							tt_opt_prism_noise
							tt_opt_usec,
							2, 0);

  mpf_set_default_prec(WIPAL_MPF_PRECISION);

  try
    {
      using wifi::frame::filter::provide_lr_synchronizer;

      typedef wifi::frame::seq_bss_tmp_id	unique_id;
      typedef pcapxx::list_of_traces<>		list_of_traces;

      wifi::addr_mapping	mapping (options.mapping);
      list_of_traces		traces1 (options.input[0]);
      list_of_traces		traces2 (options.input[1]);

      traces1.expect(pcapxx::descriptor<>::PRISM_HEADER);
      traces2.expect(pcapxx::descriptor<>::PRISM_HEADER);

      sync_unique_frames_dumper	dump (options);

      provide_lr_synchronizer<unique_id, prism::header,
			      wifi::frame::filter::fast_intersector>
	(traces1.begin(), traces1.end(),
	 traces2.begin(), traces2.end(),
	 mapping, dump,
	 options.prism_noise_means_error,
	 options.endianness,
	 options.ref_blacklist);
    }
  catch (const std::exception& e)
    {
      std::cerr << "ERROR: " << e.what() << '.' << std::endl;
      return EXIT_FAILURE;
    }
}
