/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */

#include <stdexcept>
#include <iostream>
#include <cstdlib>

#include "config.hh"

#include <wipal/tool/tt_options.hh>
#include <wipal/pcap/list_of_traces.hh>
#include <wipal/phy/prism_header.hh>
#include <wipal/phy/rtap_header.hh>

namespace
{

  pcapxx::descriptor<>::link_type
  check_linktype(const pcapxx::list_of_traces<>& l,
		 const pcapxx::list_of_traces<>& r,
		 bool				 warn)
  {
    typedef pcapxx::descriptor<>::link_type	link_type;

#define P pcapxx::descriptor<>::PRISM_HEADER

    const link_type lt = l.empty() ? P : l.iterable(0).linktype();
    const link_type rt = r.empty() ? P : r.iterable(0).linktype();

#undef P

    l.expect(lt);
    r.expect(rt);

    if (warn and lt != rt)
      std::cerr << "WARNING: " << l.file_name() << " and " << r.file_name()
		<< " have different link types (" << lt << " and " << rt
		<< "), using " << lt << '.' << std::endl;

    return lt;
  }

  bool
  eq_default(const pcapxx::frame_descriptor& l,
	     const pcapxx::frame_descriptor& r)
  {
    return l == r;
  }

  bool
  eq_802_from_prism(const pcapxx::frame_descriptor& l,
		    const pcapxx::frame_descriptor& r)
  {
    const prism::header* const lh =
      reinterpret_cast<const prism::header*> (l.bytes().get());

    const prism::header* const rh =
      reinterpret_cast<const prism::header*> (r.bytes().get());

    return eq_80211(lh, l.pcap_header()->caplen, rh, r.pcap_header()->caplen);
  }

  bool
  eq_802_from_rtap(const pcapxx::frame_descriptor& l,
		   const pcapxx::frame_descriptor& r)
  {
    const rtap::header* const lh =
      reinterpret_cast<const rtap::header*> (l.bytes().get());

    const rtap::header* const rh =
      reinterpret_cast<const rtap::header*> (r.bytes().get());

    return eq_80211(lh, l.pcap_header()->caplen, rh, r.pcap_header()->caplen);
  }

  bool
  eq_bytes(const pcapxx::frame_descriptor& l,
	   const pcapxx::frame_descriptor& r)
  {
    return
      0 == memcmp(l.bytes().get(), r.bytes().get(),
		  std::min(l.pcap_header()->caplen, r.pcap_header()->caplen));
  }

  bool
  eq_prism_time(const pcapxx::frame_descriptor& l,
		const pcapxx::frame_descriptor& r)
  {
    const pcapxx::pkthdr* const lp = l.pcap_header().get();
    const pcapxx::pkthdr* const rp = r.pcap_header().get();

    const prism::header* const	lh =
      reinterpret_cast<const prism::header*> (l.bytes().get());

    const prism::header* const	rh =
      reinterpret_cast<const prism::header*> (r.bytes().get());

    return eq_time_and_80211(lh, lp->caplen, lp->swapped,
			     rh, rp->caplen, rp->swapped,
			     phy::eq_time_default_precision,
			     true);
  }

  bool
  eq_rtap_time(const pcapxx::frame_descriptor& l,
	       const pcapxx::frame_descriptor& r)
  {
    const pcapxx::pkthdr* const lp = l.pcap_header().get();
    const pcapxx::pkthdr* const rp = r.pcap_header().get();

    const rtap::header* const	lh =
      reinterpret_cast<const rtap::header*> (l.bytes().get());

    const rtap::header* const	rh =
      reinterpret_cast<const rtap::header*> (r.bytes().get());

    return eq_time_and_80211(lh, lp->caplen, rh, rp->caplen);
  }

  template <class I1, class I2, class Eq>
  bool
  cmp(I1& l, const I1& lend,
      I2& r, const I2& rend,
      Eq eq)
  {
    while (l != lend and r != rend and eq(*l, *r))
      {
	++l;
	++r;
      }
    return l == lend and r == rend;
  }

} // End of anonymous namespace.

int
main(int argc, char *const argv[])
{
  const tool::opt::options options =
    tool::opt::parse_args(argc, argv,
			  tt_opt_quiet
			  tt_opt_802cmp
			  tt_opt_bytescmp
			  tt_opt_defaultcmp
			  tt_opt_timecmp,
			  2, 0);

  try
    {
      typedef pcapxx::list_of_traces<>		list_of_traces;
      typedef pcapxx::descriptor<>::link_type	link_type;
      typedef list_of_traces::const_iterator	iterator;

      list_of_traces	lhs (options.input[0]);
      list_of_traces	rhs (options.input[1]);
      const link_type	type = check_linktype(lhs, rhs, options.print_errors);

      iterator		l = lhs.begin();
      iterator		r = rhs.begin();
      const iterator	lend = lhs.end();
      const iterator	rend = rhs.end();

      bool		eq (false);

      switch (options.cmp)
	{
	case tool::opt::options::cmp_default:
	  eq = cmp(l, lend, r, rend, eq_default);
	  break;

	case tool::opt::options::cmp_802:
	  switch (type)
	    {
	    case pcapxx::descriptor<>::PRISM_HEADER:
	      eq = cmp(l, lend, r, rend, eq_802_from_prism);
	      break;

	    case pcapxx::descriptor<>::RADIOTAP_HEADER:
	      eq = cmp(l, lend, r, rend, eq_802_from_rtap);
	      break;

	    case pcapxx::descriptor<>::IEEE802_11:
	      eq = cmp(l, lend, r, rend, eq_bytes);
	      break;

	    default:
	      throw std::invalid_argument ("Invalid link types: "
					   "expected IEEE 802.11 frames, "
					   "Prism headers, or "
					   "radiotap headers");
	    }
	  break;

	case tool::opt::options::cmp_bytes:
	  eq = cmp(l, lend, r, rend, eq_bytes);
	  break;

	case tool::opt::options::cmp_time:
	  switch (type)
	    {
	    case pcapxx::descriptor<>::PRISM_HEADER:
	      eq = cmp(l, lend, r, rend, eq_prism_time);
	      break;

	    case pcapxx::descriptor<>::RADIOTAP_HEADER:
	      eq = cmp(l, lend, r, rend, eq_rtap_time);
	      break;

	    default:
	      throw std::invalid_argument ("Invalid link types: expected "
					   "Prism or radiotap headers");
	    }
	  break;

	default:
	  throw std::logic_error ("Unknown logic error.  Report to "
				  PACKAGE_BUGREPORT);
	}

      if (options.print_summary)
	{
	  std::cout << argv[0] << ": "
		    << lhs.file_name() << " and " << rhs.file_name() << ' ';
	  switch (eq)
	    {
	    case true:
	      std::cout << "are equals (after endianness adjustments.)";
	      break;
	    case false:
	      std::cout << "differs, frame "
			<< (l == lend ?
			    r->pcap_header()->id:
			    l->pcap_header()->id)
			<< '.';
	      break;
	    }
	  std::cout << std::endl;
	}
      return eq ? EXIT_SUCCESS : EXIT_FAILURE;
    }
  catch (std::exception& e)
    {
      std::cerr << "ERROR: " << e.what() << '.' << std::endl;
      return EXIT_FAILURE;
    }
}
