/*
 * WiPal - A library and a set of tools to manipulate wireless traces.
 * Copyright (C) 2007  Universite Pierre et Marie Curie - Paris 6
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 *
 * Author: Thomas Claveirole <thomas.claveirole@lip6.fr>
 */

#include <utility>
#include <stdexcept>
#include <iostream>
#include <cstdlib>
#include <map>

#include <boost/foreach.hpp>

#include "config.hh"

#include <wipal/tool/tt_options.hh>
#include <wipal/pcap/list_of_traces.hh>
#include <wipal/wifi/frame/dissector/dissector.hh>
#include <wipal/wifi/frame/dissector/default_hooks.hh>
#include <wipal/wifi/frame/frame.hh>
#include <wipal/wifi/frame/mgt.hh>
#include <wipal/wifi/frame/filter/non_noisy_prism.hh>

namespace
{

  std::pair<size_t, const void*>
  get_ieee80211_frame(const pcapxx::frame_descriptor& d, bool has_prism)
  {
    size_t	caplen = d.pcap_header()->caplen;
    const void*	frame;

    if (has_prism)
      {
	if (caplen < sizeof (prism::header))
	  return std::pair<size_t, const void*>(0, 0);

	caplen -= sizeof (prism::header);
	frame = reinterpret_cast<const prism::header*> (d.bytes().get()) + 1;
      }
    else
      frame = d.bytes().get();
    return std::make_pair(caplen, frame);
  }

  struct data_src: public wifi::frame::dissector_default_hooks
  {

    data_src(): check_ (false), src_ (0)
    {
    }

    void
    data_hook(const void* f, size_t)
    {
      // Ignore retransmitted frames.
      check_ = not wifi::frame::control_flag(3, f);
    }

    void
    addr_hook(const void*, size_t, unsigned i, const wifi::addr& a)
    {
      if (check_ and 2 == i)
	src_ = &a;
    }

    const wifi::addr*
    src() const
    {
      return src_;
    }

  private:
    bool		check_;
    const wifi::addr*	src_;

  }; // End of data_src.

  template <class T>
  bool
  find_data_dups(T& t, bool has_prism, bool print_summary)
  {
    using wifi::frame::dissector;

    typedef pcapxx::frame_descriptor			frame_descriptor;
    typedef std::map<wifi::addr, frame_descriptor>	last_data_map;
    typedef typename T::const_iterator			iterator;
    typedef std::pair<size_t, const void*>		ieee80211_frame;

    last_data_map	map;
    bool		success (true);
    const iterator	e = t.end();

    for (iterator i = t.begin(); i != e; ++i)
      {
	const ieee80211_frame	  frame = get_ieee80211_frame(*i, has_prism);
	const dissector<data_src> dissect (frame.second, frame.first);

	if (dissect.src())
	  {
	    const std::pair<typename last_data_map::iterator, bool> r =
	      map.insert(std::make_pair(*dissect.src(), *i));

	    if (r.second)
	      continue;

	    const ieee80211_frame last = get_ieee80211_frame(r.first->second,
							     has_prism);
	    if (0 == memcmp(frame.second, last.second,
			    std::min(frame.first, last.first)))
	      {
		success = false;
		if (print_summary)
		  std::cout << *i << '\t' << r.first->second << '\n';

	      }
	    r.first->second = *i;
	  }
      }
    return success;
  }

} // End of anonymous namespace.

int
main(int argc, char *const argv[])
{
  const tool::opt::options options =
    tool::opt::parse_args(argc, argv,
			  tt_opt_colheads
			  tt_opt_prism_noise
			  tt_opt_quiet,
			  tool::opt::options::arbitrary_number_of_inputs, 0);
  bool success (true);

  try
    {
      typedef pcapxx::list_of_traces<>	list_of_traces;

      BOOST_FOREACH(list_of_traces t, options.input)
	{
	  if (options.print_column_headers)
	    std::cout << t.file_name() << ":\n"
		         "Frame\tIs a duplicate of\n"
		         "=====\t=================" << std::endl;

	  if (t.empty())
	    continue;

	  if (options.prism_noise_means_error)
	    t.expect(pcapxx::descriptor<>::PRISM_HEADER);

	  switch (const int lt = t.iterable(0).linktype())
	    {
	    case pcapxx::descriptor<>::PRISM_HEADER:
	      if (options.prism_noise_means_error)
		{
		  typedef list_of_traces::const_iterator		i;
		  typedef wifi::frame::filter::non_noisy_prism<i>	nnp;

		  nnp	n (t.begin(), t.end());

		  success = (find_data_dups(n, true, options.print_summary)
			     and success);
		}
	      else
		success = (find_data_dups(t, true, options.print_summary)
			   and success);
	      break;

	    case pcapxx::descriptor<>::IEEE802_11:
	      success = (find_data_dups(t, false, options.print_summary)
			 and success);
	      break;

	    default:
	      success = false;
	      std::cerr << "WARNING: Invalid link type " << lt
			<< " for " << t.file_name()
			<< ".  Expected IEEE 802.11 or Prism headers.  "
		           "Ignoring trace." << std::endl;
	      break;
	    }
	}
    }
  catch (std::exception& e)
    {
      success = false;
      std::cerr << "ERROR: " << e.what() << '.' << std::endl;
    }

  return success ? EXIT_SUCCESS : EXIT_FAILURE;
}
